/*------------------------------------------------------------------*/
/*  YHF_Tone.cpp                                                    */
/*                                                                  */
/* DL4YHF's tiny tone signal generator.                             */
/*                                                                  */
/*       Originally just a simple routine to generate tones         */
/*       with the PC's speaker, now adapted for Windoze+Soundcard.  */
/*                                                                  */
/* Copyright (c) 2001 by Wolfgang Buescher                          */
/*                                                                  */
/* Autor: Wolfgang Buescher, (www.qsl.net/DL4YHF)                   */
/*                                                                  */
/* Revision history:                                                */
/*  10.Sep.2001 :                                                   */
/*        Recycled from another project.                            */
/*  2007-05-11 :  Merged GS_sound_midi.c    into  YHF_Tone.cpp  .   */
/*                                                                  */
/*------------------------------------------------------------------*/

// Notes:
//
//  When using DevCpp, select "Project".."Project Options".."PARAMETERS"(!)..
//   .. "Add library or object" , and find your way to "libwinmm.a" .
//   If you don't hear anything, open the soundcard's volume control panel,
//   enable the "MIDI-Synth" source and/or crank up its volume .
//
//  This module produces simple "beep-like" sounds, sent as short MIDI messages
//  to the default MIDI Out device.


#include <windows.h>
#include <stdio.h>
#include <mmsystem.h>
#include <math.h>     // believe it or not, we need a logarithm in here !

#include "YHF_Tone.h"  // header for THIS module

#define  C_MIDI_CHANNEL_BEEPER 0x90  /* MIDI "status byte" for "Channel 1 Note ON" (!) */

HMIDIOUT YHF_hMIDIOut = NULL;
BYTE     YHF_bMIDIPlayingSimpleNote = 0;

int YHF_InitTones(void)  // Must be called once to initialize the sound output.
{
  if( YHF_hMIDIOut != NULL )
   { midiOutClose(YHF_hMIDIOut);
     YHF_hMIDIOut = NULL; // make sure no-one talks to an invalid handle ;-)
   }

  if( midiOutOpen(&YHF_hMIDIOut, (UINT)-1, 0, 0, CALLBACK_NULL) != MMSYSERR_NOERROR )
   { // bad luck, the default midi output device won't talk to us.
     YHF_hMIDIOut = NULL;  // make sure no-one talks to an invalid handle ;-)
   }
  return YHF_hMIDIOut != NULL;

} // end YHF_InitTones()

void YHF_CloseTones(void)  // Must be called once on exit, to close handles etc.
{
  if( YHF_hMIDIOut != NULL )
   { midiOutClose(YHF_hMIDIOut);
     YHF_hMIDIOut = NULL; // make sure no-one talks to an invalid handle ;-)
   }
} // end YHF_CloseTones()

void YHF_SendMIDIEvent( BYTE bStatus, BYTE bData1, BYTE bData2 )
{
  union   // union for packing up to 3 MIDI bytes into a single DWORD
   { DWORD dwData;
     BYTE bData[4];
   } u;

  // Construct the MIDI message.
  u.bData[0] = bStatus;  // MIDI status byte
  u.bData[1] = bData1;   // first MIDI data byte
  u.bData[2] = bData2;   // second MIDI data byte
  u.bData[3] = 0;        // fourth byte shall always be ZERO

  // Send the message. THAT'S ALL .. unbelievably simple for a Win API call !
  if( YHF_hMIDIOut != NULL )
   { midiOutShortMsg( YHF_hMIDIOut, u.dwData);
   }
} // end YHF_SendMIDIEvent()

BYTE YHF_FrequencyToMidiNote( int freq )
{
  // How to convert a frequency [Hz] into a MIDI note number ?
  // The MIDI frequency range starts at MIDI note 0, C = 8.1758 Hz,
  // and extends above MIDI note 127, G = 12543.875 Hz.
  // From http://logic-users.org/forums/L-OT/295
  // Each note's frequency is 2^(1/12) times of the previous note.
  //  freq = 440 * 2^((n-69)/12)
  //  n = 69 + 12*log(freq/440)/log(2)
  // Proof:
  //   69 + 12*log(8.1758/440)/log(2)    =: 2.29614e-06
  //   69 + 12*log(12543.875/440)/log(2) =: 127  (bingo)
  double dblNote = 69.0 + 12.0 * log((double)freq/440.0) / log(2);
  if( dblNote<0.0 )  dblNote=0.0;
  if( dblNote>255.0) dblNote=255.0;
  return (BYTE)(dblNote+0.499);
} // end YHF_FrequencyToMidiNote()

void YHF_ChangeInstrument( int gm_prog_nr )
  // [in]: gm_prog_nr = General MIDI "Program Number",
  //       see http://de.wikipedia.org/wiki/General_MIDI
  //  1 = Acoustic Grand Piano,   17 = Organ,   74 = Flute, etc etc etc .
  // (for portability, use YHF_INSTRUMENT_PIANO,YHF_INSTRUMENT_ORGAN, etc )
{
  // Details (much better than those from Wikipedia) could be found here:
  //          http://www.borg.com/~jglatt/tutr/gm.htm
  YHF_SendMIDIEvent( 0xC0/*"Program Change"*/, gm_prog_nr, 0 );
} // end YHF_ChangeInstrument()

void YHF_PlaySimpleToneOnMidi( int freq )  // freq in Hertz; 0=turn beep off
{
  BYTE bNoteNr = 0;
  // Send a MIDI "note" message :
  // - MIDI "status" : 0x90 for "Play note on MIDI channel 1", etc
  //   (a programmer wouldn't call this "status", it's in fact a "command")
  // - 1st MIDI data byte = note number.
  // - 2nd MIDI data byte = note velocity.
  if( freq>0 )  // find a suitable note-number for this frequency:
   { bNoteNr = YHF_FrequencyToMidiNote( freq );
   }
  // If the "beeper" MIDI channel is already playing a note, stop it:
  if( YHF_bMIDIPlayingSimpleNote>0 )
   {
     YHF_SendMIDIEvent( C_MIDI_CHANNEL_BEEPER, YHF_bMIDIPlayingSimpleNote, 0 );
     YHF_bMIDIPlayingSimpleNote = 0;
   }
  if( bNoteNr!=0 )  // start playing a new note ?
   { YHF_bMIDIPlayingSimpleNote = bNoteNr;
     YHF_SendMIDIEvent( C_MIDI_CHANNEL_BEEPER, YHF_bMIDIPlayingSimpleNote, 127 );
   }
} // end YHF_PlaySimpleToneOnMidi()


//---------------------------------------------------------------------------
void YHF_GenerateTone(int iAudioDevice, int iFrequencyHz)
  /* Starts the output of a tone with the PC speaker
   *  or -if present- a soundcard compatible audio device.
   *
   * WARNING: CALLS A NASTY PORT ACCESS ROUTINE IN THE MAIN FORM,
   *          AND MUST NOT BE CALLED FROM ANY THREAD !!!
   *          THIS IS DEFINITELY A WAY HOW IT SHALL NOT BE DONE
   *          IN A WELL-BEHAVING WINDOZE PROGRAM.
   */
{
  long timer_value;
  if( (iAudioDevice<=-2) || (iAudioDevice==YHF_TONE_AUDIO_DEVICE_SPKR) )
   {  // use the PC SPEAKER to produce a single tone...
    if (iFrequencyHz>0)
     { // start sound output,
       // using the PC's programmable interval timer, counter 2...
      timer_value = 1193180 / iFrequencyHz;
      CPU_outp(0x0043, 0xB6);  // control register: counter #2, mode=3
      CPU_outp(0x0042, (BYTE)(timer_value & 0x00FF) );
      CPU_outp(0x0042, (BYTE)(timer_value >> 8)     );
      // now enable the sound output by setting the counter's GATE input:
      CPU_outp(0x0061, CPU_inp(0x0061) | (BYTE)0x03 );
     }
    else // do not START but STOP a tone:
     {
      // disable the sound output by resetting the counter's GATE input:
      CPU_outp(0x0061, CPU_inp(0x0061) & (BYTE)0xFC );
     }
   }
  else // do not use the PC's internal speaker, but the soundcard's synthesizer
  switch( iAudioDevice )
   { case YHF_TONE_AUDIO_DEVICE_NONE: /* no output device enabled , silence  */
        break;
     case YHF_TONE_AUDIO_DEVICE_SPKR: /* this is the INTERNAL "PC-Speaker"   */
        break;  // handled above, with negative device numbers for compatibility
     case YHF_TONE_AUDIO_DEVICE_MIDI: /* this is the midi-driven Synthesizer */
        YHF_PlaySimpleToneOnMidi( iFrequencyHz );
        break;
   } // end switch( iAudioDevice )
} // end ..GenerateSoundEffects()
